#version 430 core

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
	mat3 TBN;
} vert;

out vec4 FragColor;

layout (location = 0) out vec4 gFragColor;
layout (location = 1) out vec2 gMotionVec;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;

uniform sampler2D diffuseTexture;
uniform sampler2D normalTexture;
uniform samplerCube shadowMap; 
uniform float far_plane;

uniform bool shadows;
uniform bool normalmapping;

uniform vec3 camera_world;

uniform mat4 modelMatrix;
uniform mat4 viewProjMatrix;
// required for motion blur
uniform mat4 prevModelMatrix;
uniform mat4 prevVPMatrix;
uniform bool isCharacter;
uniform bool isChild;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
} pointL;

// array of offset direction for sampling
vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1, 1,  1), 
   vec3(1, 1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1, 1,  0),
   vec3(1, 0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1, 0, -1),
   vec3(0, 1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0, 1, -1)
);

float ShadowCalculation(vec3 fragPos)
{
	vec3 fragToLight = fragPos - pointL.position;
    float currentDepth = length(fragToLight);   
    float shadow = 0.0;
    float bias = 0.15;
    int samples = 20;
    float viewDistance = length(camera_world - fragPos);
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;
    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(shadowMap, fragToLight + gridSamplingDisk[i] * diskRadius).r;
        closestDepth *= far_plane;   // undo mapping [0;1]
        if(currentDepth - bias > closestDepth)
            shadow += 1.0;
    }
    shadow /= float(samples);
    return shadow;
}

void main() {
    vec3 position_world = (modelMatrix * vec4(vert.position_world, 1.0)).xyz;

	vec3 normal;
	if (!normalmapping)
	{
	normal = normalize(vert.normal_world);

	} else 
	{
	normal = texture(normalTexture, vert.uv).rgb;
	normal = normalize(normal * 2.0 - 1.0);
	normal = normalize(vert.TBN * normal);

	}

	vec3 color = texture(diffuseTexture, vert.uv).rgb;
    vec3 lightColor = pointL.color;
    // ambient
    vec3 ambient = materialCoefficients.x * color;
    // diffuse
    vec3 lightDir = normalize(pointL.position - position_world);
    float diff = max(dot(lightDir, normal), 0.0);
    vec3 diffuse = diff * lightColor;
    // specular
    vec3 viewDir = normalize(camera_world - position_world);
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = 0.0;
    vec3 halfwayDir = normalize(lightDir + viewDir);  
    spec = pow(max(dot(normal, halfwayDir), 0.0), 64.0);
    vec3 specular = spec * lightColor;    
    // calculate shadow
    float shadow = shadows ? ShadowCalculation(position_world) : 0.0;                      
    vec3 lighting = (ambient + (1.0 - shadow) * (diffuse + specular)) * color;    
    
    FragColor = vec4(lighting, 1.0);

    // calculate motion vector for each pixel
    vec4 previous_position = prevVPMatrix * prevModelMatrix * vec4(vert.position_world, 1.0);
    previous_position /= previous_position.w;

    vec4 current_position = viewProjMatrix * modelMatrix * vec4(vert.position_world, 1.0);
    current_position /= current_position.w;

    gFragColor = vec4(lighting, 1.0);
    gMotionVec = isCharacter || isChild ? vec2(0, 0) : (previous_position.xy - current_position.xy)/2.f;
}

